import io
import json
from unittest import TestCase

from acnportal import acnsim
from acnportal.algorithms import BaseAlgorithm, UncontrolledCharging
from acnportal.acnsim.base import ErrorAllWrapper
from tests.serialization_extensions import NamedEvent, DefaultNamedEvent
from tests.serialization_extensions import SetAttrEvent, BatteryListEvent

import os
import numpy as np
from copy import deepcopy
from datetime import datetime


class TestJSONIO(TestCase):
    @classmethod
    def setUpClass(cls):
        # Make instance of each class in registry.
        # Battery
        cls.battery1 = acnsim.Battery(100, 50, 20)
        cls.battery1._current_charging_power = 10

        # Linear2StageBattery
        cls.battery2 = acnsim.Linear2StageBattery(100, 50, 20)
        cls.battery2._current_charging_power = 10
        cls.battery2._noise_level = 0.1
        cls.battery2._transition_soc = 0.85

        # EVs
        staying_time = 10
        ev1_arrival = 10
        cls.ev1 = acnsim.EV(
            ev1_arrival,
            ev1_arrival + staying_time,
            30,
            "PS-001",
            "EV-001",
            deepcopy(cls.battery1),
            estimated_departure=35,
        )
        cls.ev1._energy_delivered = 0.05
        cls.ev1._current_charging_rate = 10

        ev2_arrival = 40
        cls.ev2 = acnsim.EV(
            ev2_arrival,
            ev2_arrival + staying_time,
            30,
            "PS-002",
            "EV-002",
            deepcopy(cls.battery2),
            estimated_departure=65,
        )
        cls.ev2._energy_delivered = 0.05
        cls.ev2._current_charging_rate = 10

        ev3_arrival = 50
        cls.ev3 = acnsim.EV(
            ev3_arrival,
            ev3_arrival + staying_time,
            30,
            "PS-003",
            "EV-003",
            deepcopy(cls.battery2),
            estimated_departure=75,
        )
        cls.ev3._energy_delivered = 0.05
        cls.ev3._current_charging_rate = 10

        # EVSEs
        cls.evse0 = acnsim.EVSE("PS-000", max_rate=32)

        cls.evse1 = acnsim.EVSE("PS-001", max_rate=32)
        cls.evse1.plugin(cls.ev1)
        cls.evse1.set_pilot(30, 220, 1)

        cls.evse2 = acnsim.DeadbandEVSE("PS-002", max_rate=32, deadband_end=4)
        cls.evse2.plugin(cls.ev2)
        cls.evse2.set_pilot(30, 220, 1)

        cls.evse3 = acnsim.FiniteRatesEVSE("PS-003", allowable_rates=[0, 8, 16, 24, 32])
        cls.evse3.plugin(cls.ev3)
        cls.evse3.set_pilot(24, 220, 1)

        # Events
        cls.event = acnsim.Event(0)
        cls.plugin_event1 = acnsim.PluginEvent(10, cls.ev1)
        cls.unplug_event = acnsim.UnplugEvent(20, cls.ev1)
        cls.recompute_event1 = acnsim.RecomputeEvent(30)
        cls.plugin_event2 = acnsim.PluginEvent(40, cls.ev2)
        cls.plugin_event3 = acnsim.PluginEvent(50, cls.ev3)
        # Modify a default attribute to check if it's loaded correctly.
        cls.recompute_event2 = acnsim.RecomputeEvent(10)
        cls.recompute_event2.event_type = "Recompute Modified"

        # EventQueue
        cls.event_queue = acnsim.EventQueue()
        cls.event_queue.add_events(
            [
                cls.event,
                cls.plugin_event1,
                cls.recompute_event1,
                cls.plugin_event2,
                cls.plugin_event3,
            ]
        )

        # Network
        cls.network = acnsim.ChargingNetwork(
            violation_tolerance=1e-3, relative_tolerance=1e-5
        )

        cls.network.register_evse(cls.evse1, 220, 30)
        cls.network.register_evse(cls.evse2, 220, 150)
        cls.network.register_evse(cls.evse3, 220, -90)
        cls.network.constraint_matrix = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        cls.network.magnitudes = np.array([32, 32, 32])
        cls.network.constraint_index = ["C1", "C2", "C3"]
        _ = cls.network._update_info_store()
        cls.empty_network = acnsim.ChargingNetwork()

        # Simulator
        cls.simulator = acnsim.Simulator(
            cls.network,
            UncontrolledCharging(),
            cls.event_queue,
            datetime(2019, 1, 1),
            verbose=False,
            store_schedule_history=True,
        )

        # Make a copy of the simulator to run
        cls.simulator_run = deepcopy(cls.simulator)
        # Do necessary unplugs.
        for evse in cls.simulator_run.network._EVSEs.values():
            if evse.ev is not None:
                evse.unplug()
        # Run simulation
        cls.simulator_run.run()

        # Make a copy of the simulator with signals
        cls.simulator_signal = deepcopy(cls.simulator)
        cls.simulator_signal.signals = {"a": [0, 1, 2], "b": [3, 4]}

        cls.simulator_hard_signal = deepcopy(cls.simulator)
        cls.simulator_hard_signal.signals = {"a": BaseAlgorithm()}

        cls.simulator_no_sch_hist = deepcopy(cls.simulator)
        cls.simulator_no_sch_hist.schedule_history = None

        # Class data used for testing.
        cls.simple_attributes = {
            "Battery": [
                "_max_power",
                "_current_charging_power",
                "_current_charge",
                "_capacity",
                "_init_charge",
            ],
            "Linear2StageBattery": [
                "_max_power",
                "_current_charging_power",
                "_current_charge",
                "_capacity",
                "_init_charge",
                "_noise_level",
                "_transition_soc",
            ],
            "EV": [
                "_arrival",
                "_departure",
                "_requested_energy",
                "_estimated_departure",
                "_session_id",
                "_station_id",
                "_energy_delivered",
                "_current_charging_rate",
            ],
            "EVSE": [
                "_station_id",
                "_max_rate",
                "_min_rate",
                "_current_pilot",
                "is_continuous",
            ],
            "DeadbandEVSE": [
                "_station_id",
                "_max_rate",
                "_current_pilot",
                "_deadband_end",
                "is_continuous",
            ],
            "FiniteRatesEVSE": [
                "_station_id",
                "_current_pilot",
                "is_continuous",
                "allowable_rates",
            ],
            "Event": ["timestamp", "event_type", "precedence"],
            "PluginEvent": ["timestamp", "event_type", "precedence"],
            "UnplugEvent": ["timestamp", "event_type", "precedence"],
            "RecomputeEvent": ["timestamp", "event_type", "precedence"],
            "EventQueue": ["_timestep"],
            "ChargingNetwork": [
                "constraint_index",
                "violation_tolerance",
                "relative_tolerance",
                "_station_ids_dict",
            ],
            "Simulator": [
                "period",
                "max_recompute",
                "verbose",
                "peak",
                "_iteration",
                "_resolve",
                "_last_schedule_update",
                "schedule_history",
            ],
        }

    def _obj_compare_helper(self, obj, attributes=None):
        obj_json = obj.to_json()
        obj_class = type(obj)
        obj_loaded = obj_class.from_json(obj_json)
        self.assertIsInstance(obj_loaded, obj_class)
        if attributes is None:
            attributes = self.simple_attributes[obj_class.__name__]
        for attribute in attributes:
            self.assertEqual(getattr(obj, attribute), getattr(obj_loaded, attribute))
        return obj_loaded

    def test_battery_json(self):
        _ = self._obj_compare_helper(self.battery1)

    def test_linear_2_stage_battery_json(self):
        _ = self._obj_compare_helper(self.battery2)

    def test_ev_json(self):
        def _load_dump_compare_helper(ev, bat_type):
            self.assertIsInstance(ev._battery, bat_type)
            ev_loaded = self._obj_compare_helper(ev)
            self.assertIsInstance(ev_loaded._battery, bat_type)
            self.assertEqual(ev_loaded._battery.__dict__, ev._battery.__dict__)

        _load_dump_compare_helper(self.ev1, acnsim.Battery)
        _load_dump_compare_helper(self.ev2, acnsim.Linear2StageBattery)
        _load_dump_compare_helper(self.ev3, acnsim.Linear2StageBattery)

    def test_evse_no_ev_json(self):
        evse_loaded = self._obj_compare_helper(self.evse0)
        self.assertEqual(getattr(evse_loaded, "_ev"), None)

    def test_evse_json(self):
        evse_loaded = self._obj_compare_helper(self.evse1)
        self.assertEqual(getattr(getattr(evse_loaded, "_ev"), "_session_id"), "EV-001")

    def test_deadband_evse_json(self):
        evse_loaded = self._obj_compare_helper(self.evse2)
        self.assertEqual(getattr(getattr(evse_loaded, "_ev"), "_session_id"), "EV-002")

    def test_finite_rates_evse_json(self):
        evse_loaded = self._obj_compare_helper(self.evse3)
        self.assertEqual(getattr(getattr(evse_loaded, "_ev"), "_session_id"), "EV-003")

    def test_event_json(self):
        _ = self._obj_compare_helper(self.event)

    def test_plugin_event_json(self):
        event_loaded = self._obj_compare_helper(self.plugin_event1)
        self.assertEqual(getattr(getattr(event_loaded, "ev"), "_session_id"), "EV-001")

    def test_unplug_event_json(self):
        event_loaded = self._obj_compare_helper(self.unplug_event)
        self.assertEqual(getattr(getattr(event_loaded, "ev"), "_session_id"), "EV-001")

    def test_recompute_event_json(self):
        _ = self._obj_compare_helper(self.recompute_event1)

    def test_altered_recompute_event_json(self):
        _ = self._obj_compare_helper(self.recompute_event2)

    def test_event_queue_json(self):
        event_queue_loaded = self._obj_compare_helper(self.event_queue)
        for (ts, event), (tsl, event_loaded) in zip(
            self.event_queue._queue, event_queue_loaded._queue
        ):
            self.assertEqual(ts, tsl)
            self.assertIsInstance(event_loaded, acnsim.Event)
            if not isinstance(event, acnsim.PluginEvent):
                self.assertEqual(type(event), type(event_loaded))
                self.assertEqual(event.__dict__, event_loaded.__dict__)
            else:
                for field in ["timestamp", "event_type", "precedence"]:
                    self.assertEqual(
                        getattr(event, field), getattr(event_loaded, field)
                    )
                self.assertEqual(
                    getattr(getattr(event_loaded, "ev"), "_session_id"),
                    getattr(getattr(event, "ev"), "_session_id"),
                )

    def test_empty_charging_network_json(self):
        empty_network_loaded = self._obj_compare_helper(self.empty_network)
        self.assertIsInstance(empty_network_loaded, acnsim.ChargingNetwork)

        network_np_fields = [
            "magnitudes",
            "_voltages",
            "_phase_angles",
            "max_pilot_signals",
            "min_pilot_signals",
            "is_continuous",
        ]
        for field in network_np_fields:
            np.testing.assert_equal(
                getattr(self.empty_network, field), getattr(empty_network_loaded, field)
            )
        extra_simple_attributes = ["constraint_matrix", "_EVSEs", "allowable_rates"]
        for attribute in extra_simple_attributes:
            self.assertEqual(
                getattr(self.empty_network, attribute),
                getattr(empty_network_loaded, attribute),
            )

    def test_charging_network_json(self):
        network_loaded = self._obj_compare_helper(self.network)
        self.assertIsInstance(network_loaded, acnsim.ChargingNetwork)

        network_np_fields = [
            "constraint_matrix",
            "magnitudes",
            "_voltages",
            "_phase_angles",
            "max_pilot_signals",
            "min_pilot_signals",
            "is_continuous",
        ]
        for field in network_np_fields:
            np.testing.assert_equal(
                getattr(self.network, field), getattr(network_loaded, field)
            )

        for (station_id, evse), (station_id_l, evse_l) in zip(
            self.network._EVSEs.items(), network_loaded._EVSEs.items()
        ):
            self.assertEqual(station_id, station_id_l)
            self.assertEqual(evse.station_id, evse_l.station_id)

        for allowable_rates_array, allowable_rates_array_loaded in zip(
            self.network.allowable_rates, network_loaded.allowable_rates
        ):
            np.testing.assert_equal(allowable_rates_array, allowable_rates_array_loaded)

    def _sim_compare_helper(self, sim):
        simulator_loaded = self._obj_compare_helper(sim)

        self.assertEqual(sim.start, simulator_loaded.start)

        self.assertIsInstance(simulator_loaded.scheduler, UncontrolledCharging)

        if simulator_loaded.signals is not None:
            self.assertEqual(sim.signals, simulator_loaded.signals)

        np.testing.assert_equal(sim.pilot_signals, simulator_loaded.pilot_signals)
        np.testing.assert_equal(sim.charging_rates, simulator_loaded.charging_rates)

        network_attrs = [
            "station_ids",
            "active_station_ids",
            "voltages",
            "phase_angles",
        ]
        for attr in network_attrs:
            self.assertEqual(
                getattr(sim.network, attr), getattr(simulator_loaded.network, attr)
            )

        for (ts, event), (tsl, event_loaded) in zip(
            sim.event_queue._queue, simulator_loaded.event_queue._queue
        ):
            self.assertEqual(ts, tsl)
            self.assertEqual(event.event_type, event_loaded.event_type)
            self.assertEqual(event.timestamp, event_loaded.timestamp)

        for event, event_loaded in zip(
            sim.event_history, simulator_loaded.event_history
        ):
            self.assertEqual(event.event_type, event_loaded.event_type)
            self.assertEqual(event.timestamp, event_loaded.timestamp)

        for station_id, ev in sim.ev_history.items():
            self.assertEqual(
                ev.session_id, simulator_loaded.ev_history[station_id].session_id
            )

    def test_init_simulator_json(self):
        self._sim_compare_helper(self.simulator)

    def test_run_simulator_json(self):
        self._sim_compare_helper(self.simulator_run)

    def test_object_equalities(self):
        # Tests that certain object equality invariants are preserved
        # after loading.
        simulator_json = self.simulator_run.to_json()
        simulator_loaded = acnsim.Simulator.from_json(simulator_json)

        plugins = filter(
            lambda x: isinstance(x, acnsim.PluginEvent), simulator_loaded.event_history
        )
        evs = list(simulator_loaded.ev_history.values())
        for plugin, ev in zip(plugins, evs):
            self.assertIs(plugin.ev, ev)

    def test_sim_signal_json(self):
        self._sim_compare_helper(self.simulator_signal)

    def test_sim_signal_warning(self):
        with self.assertWarns(UserWarning):
            self._sim_compare_helper(self.simulator_hard_signal)

    def test_sim_no_sch_hist(self):
        self._sim_compare_helper(self.simulator_no_sch_hist)

    def test_acnportal_version_inequality(self):
        with self.assertWarns(UserWarning):
            acnsim.Event.from_json(
                os.path.join(os.path.dirname(__file__), "old_version.json")
            )

    def test_numpy_version_inequality(self):
        with self.assertWarns(UserWarning):
            acnsim.Event.from_json(
                os.path.join(os.path.dirname(__file__), "old_dependencies.json")
            )


class TestExtObjJSONIO(TestJSONIO):
    battery1 = None
    event_queue = None
    simple_attributes = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.set_battery_event = SetAttrEvent(5)
        cls.set_battery_event.set_extra_attr(cls.battery1)

        cls.event_queue.add_event(cls.set_battery_event)

        cls.simple_attributes["DefaultNamedEvent"] = [
            "timestamp",
            "event_type",
            "precedence",
        ]
        cls.simple_attributes["SetAttrEvent"] = [
            "timestamp",
            "event_type",
            "precedence",
            "extra_attr",
        ]
        cls.simple_attributes["BatteryListEvent"] = [
            "timestamp",
            "event_type",
            "precedence",
        ]

    def test_named_event_json(self):
        named_event = NamedEvent(0, "my_event")
        with self.assertWarns(UserWarning):
            named_event_json = named_event.to_json()
        with self.assertRaises(TypeError):
            NamedEvent.from_json(named_event_json)

    def test_default_named_event_json(self):
        default_named_event = DefaultNamedEvent(5, "def_event")
        with self.assertWarns(UserWarning):
            default_named_event_loaded = self._obj_compare_helper(default_named_event)
        self.assertEqual(default_named_event_loaded.name, "my_event")

    def _obj_compare_helper_warning(self, obj, attributes=None):
        with self.assertWarns(UserWarning):
            obj_json = obj.to_json()
        obj_class = type(obj)
        with self.assertWarns(UserWarning):
            obj_loaded = obj_class.from_json(obj_json)
        self.assertIsInstance(obj_loaded, obj_class)
        if attributes is None:
            attributes = self.simple_attributes[obj_class.__name__]
        for attribute in attributes:
            self.assertEqual(getattr(obj, attribute), getattr(obj_loaded, attribute))
        return obj_loaded

    def test_set_named_event_json(self):
        set_named_event = SetAttrEvent(5)
        set_named_event.set_extra_attr("set_event")
        _ = self._obj_compare_helper_warning(set_named_event)

    def test_set_list_event_json(self):
        set_list_event = SetAttrEvent(5)
        set_list_event.set_extra_attr(["set_event1", "set_event2"])
        _ = self._obj_compare_helper_warning(set_list_event)

    def test_set_battery_event_json(self):
        set_battery_event_loaded = self._obj_compare_helper_warning(
            self.set_battery_event, self.simple_attributes["Event"]
        )
        self.assertEqual(
            self.battery1.__dict__, set_battery_event_loaded.extra_attr.__dict__
        )

    def test_set_np_event_json(self):
        set_np_event = SetAttrEvent(5)
        set_np_event.set_extra_attr(np.zeros((2, 2)))
        set_np_event_loaded = self._obj_compare_helper_warning(
            set_np_event, self.simple_attributes["Event"]
        )
        self.assertIsInstance(set_np_event_loaded.extra_attr, ErrorAllWrapper)
        self.assertEqual(
            set_np_event_loaded.extra_attr.data, "array([[0., 0.],\n       [0., 0.]])"
        )

    def test_battery_list_event_json(self):
        battery_list_event = BatteryListEvent(5, [self.battery1, self.battery2])
        battery_list_event_loaded = self._obj_compare_helper(
            battery_list_event, self.simple_attributes["Event"]
        )
        for battery, battery_loaded in zip(
            battery_list_event.battery_list, battery_list_event_loaded.battery_list
        ):
            self.assertEqual(battery.__dict__, battery_loaded.__dict__)

    def test_event_queue_json(self):
        event_queue_loaded = self._obj_compare_helper_warning(self.event_queue)
        for (ts, event), (tsl, event_loaded) in zip(
            self.event_queue._queue, event_queue_loaded._queue
        ):
            self.assertEqual(ts, tsl)
            self.assertIsInstance(event_loaded, acnsim.Event)
            if isinstance(event_loaded, acnsim.PluginEvent):
                for field in ["timestamp", "event_type", "precedence"]:
                    self.assertEqual(
                        getattr(event, field), getattr(event_loaded, field)
                    )
                self.assertEqual(
                    getattr(getattr(event_loaded, "ev"), "_session_id"),
                    getattr(getattr(event, "ev"), "_session_id"),
                )
            elif isinstance(event_loaded, SetAttrEvent):
                for field in ["timestamp", "event_type", "precedence"]:
                    self.assertEqual(
                        getattr(event, field), getattr(event_loaded, field)
                    )
                self.assertEqual(
                    getattr(event_loaded, "extra_attr").__dict__,
                    getattr(event, "extra_attr").__dict__,
                )
            else:
                self.assertEqual(type(event), type(event_loaded))
                self.assertEqual(event.__dict__, event_loaded.__dict__)

    def test_init_simulator_json(self):
        with self.assertWarns(UserWarning):
            self._sim_compare_helper(self.simulator)


class TestJSONIOTypes(TestJSONIO):
    @classmethod
    def setUpClass(cls):
        cls.battery = acnsim.Battery(100, 50, 20)

    def test_to_json_string(self):
        battery_json_string = self.battery.to_json()
        self.assertIsInstance(battery_json_string, str)
        battery_loaded = acnsim.Battery.from_json(battery_json_string)
        self.assertEqual(self.battery.__dict__, battery_loaded.__dict__)

    def test_to_json_filepath(self):
        filepath = os.path.join(os.path.dirname(__file__), "battery_test_filepath.json")
        battery_json_filepath = self.battery.to_json(filepath)
        self.assertIsNone(battery_json_filepath)
        battery_loaded = acnsim.Battery.from_json(filepath)
        self.assertEqual(self.battery.__dict__, battery_loaded.__dict__)
        # Clear the file so that this test doesn't cause the package
        # state to change.
        with open(filepath, "w") as file_handle:
            json.dump({}, file_handle)
            file_handle.write("\n")

    def test_to_json_file_handle(self):
        filepath = os.path.join(
            os.path.dirname(__file__), "battery_test_file_handle.json"
        )
        file_handle = open(filepath, "w")
        battery_json_filepath = self.battery.to_json(file_handle)
        file_handle.close()
        self.assertIsNone(battery_json_filepath)
        file_handle = open(filepath)
        battery_loaded = acnsim.Battery.from_json(file_handle)
        file_handle.close()
        self.assertEqual(self.battery.__dict__, battery_loaded.__dict__)
        # Clear the file so that this test doesn't cause the package
        # state to change.
        with open(filepath, "w") as file_handle:
            json.dump({}, file_handle)
            file_handle.write("\n")

    def test_to_json_str_io(self):
        output = io.StringIO()
        battery_json_string = self.battery.to_json(output)
        self.assertIsNone(battery_json_string)
        self.assertIsInstance(output.getvalue(), str)
        output_str = output.getvalue()
        output.close()
        input_str_io = io.StringIO(output_str)
        battery_loaded = acnsim.Battery.from_json(input_str_io)
        self.assertEqual(self.battery.__dict__, battery_loaded.__dict__)


class TestLegacyObjJSONInput(TestCase):
    def test_legacy_unplug_event_json(self) -> None:
        """ Tests that UnplugEvents from <0.2.2 can be loaded.

        In acnportal v0.2.2, UnplugEvent had session_id and station_id attributes
        instead of an ev attribute.

        Returns:
            None
        """
        with self.assertWarns(UserWarning):
            unplug_loaded: acnsim.UnplugEvent = acnsim.UnplugEvent.from_json(
                os.path.join(os.path.dirname(__file__), "old_unplug.json")
            )
        self.assertIsInstance(unplug_loaded, acnsim.UnplugEvent)
        self.assertEqual(unplug_loaded.event_type, "Unplug")
        self.assertEqual(unplug_loaded.timestamp, 11)
        self.assertEqual(unplug_loaded.precedence, 0)

        # Check that UnplugEvent's EV is partially loaded
        self.assertEqual(getattr(getattr(unplug_loaded, "ev"), "_session_id"), "EV-001")
        self.assertEqual(getattr(getattr(unplug_loaded, "ev"), "_station_id"), "PS-001")

        for attribute in [
            "arrival",
            "departure",
            "requested_energy",
            "estimated_departure",
            "battery",
            "energy_delivered",
            "current_charging_rate",
        ]:
            with self.assertRaises(AttributeError):
                getattr(unplug_loaded.ev, attribute)



from unittest import TestCase
from unittest.mock import Mock, create_autospec

import numpy as np
import pandas as pd

from acnportal.acnsim import Simulator, Interface
from acnportal.acnsim.network import ChargingNetwork, Current
from acnportal.algorithms import BaseAlgorithm, UncontrolledCharging
from acnportal.acnsim.events import EventQueue, Event
from datetime import datetime
from acnportal.acnsim.models import EVSE


class TestSimulator(TestCase):
    def setUp(self):
        start = Mock(datetime)
        network = ChargingNetwork()
        evse1 = EVSE("PS-001", max_rate=32)
        network.register_evse(evse1, 240, 0)
        evse2 = EVSE("PS-002", max_rate=32)
        network.register_evse(evse2, 240, 0)
        evse3 = EVSE("PS-003", max_rate=32)
        network.register_evse(evse3, 240, 0)
        self.scheduler = BaseAlgorithm()
        self.scheduler.max_recompute = None
        events = EventQueue(events=[Event(1), Event(2)])
        self.simulator = Simulator(network, self.scheduler, events, start)
        self.simulator_no_schedule = Simulator(network, None, events, start)

    def test_no_schedule_init(self):
        self.assertIsNone(self.simulator_no_schedule.scheduler)

    def test_no_schedule_error(self):
        with self.assertRaises(TypeError):
            self.simulator_no_schedule.run()

    def test_correct_on_init_pilot_signals(self):
        np.testing.assert_allclose(
            self.simulator.pilot_signals,
            np.zeros(
                (
                    len(self.simulator.network.station_ids),
                    self.simulator.event_queue.get_last_timestamp() + 1,
                )
            ),
        )

    def test_correct_on_init_charging_rates(self):
        np.testing.assert_allclose(
            self.simulator.charging_rates,
            np.zeros(
                (
                    len(self.simulator.network.station_ids),
                    self.simulator.event_queue.get_last_timestamp() + 1,
                )
            ),
        )

    def test_correct_on_init_scheduler_data(self):
        self.assertIsInstance(self.simulator.scheduler, BaseAlgorithm)
        self.assertIsInstance(self.scheduler.interface, Interface)
        self.assertIsNone(self.simulator.max_recompute)

    def test_update_schedules_not_in_network(self):
        new_schedule = {"PS-001": [24, 16], "PS-004": [16, 24]}
        with self.assertRaises(KeyError):
            self.simulator._update_schedules(new_schedule)

    def test_update_schedules_valid_schedule(self):
        new_schedule = {"PS-001": [24, 16], "PS-002": [16, 24]}
        self.simulator._update_schedules(new_schedule)
        np.testing.assert_allclose(
            self.simulator.pilot_signals[:, :2], np.array([[24, 16], [16, 24], [0, 0]])
        )

    def test_index_of_evse_error(self):
        with self.assertRaises(KeyError):
            _ = self.simulator.index_of_evse("PS-004")

    def test_index_of_evse(self):
        idx = self.simulator.index_of_evse("PS-002")
        self.assertEqual(idx, 1)

    def test_pilot_signals_as_df(self):
        self.simulator.pilot_signals = np.array([[1, 2], [3, 4], [5, 6]])
        outframe = self.simulator.pilot_signals_as_df()
        pd.testing.assert_frame_equal(
            outframe,
            pd.DataFrame(
                np.array([[1, 3, 5], [2, 4, 6]]), columns=["PS-001", "PS-002", "PS-003"]
            ),
        )

    def test_charging_rates_as_df(self):
        self.simulator.charging_rates = np.array([[1.1, 2.1], [3.1, 4.1], [5.1, 6.1]])
        outframe = self.simulator.charging_rates_as_df()
        pd.testing.assert_frame_equal(
            outframe,
            pd.DataFrame(
                np.array([[1.1, 3.1, 5.1], [2.1, 4.1, 6.1]]),
                columns=["PS-001", "PS-002", "PS-003"],
            ),
        )

    def test_update_scheduler(self):
        new_scheduler = UncontrolledCharging()
        self.assertIsNone(new_scheduler._interface)
        self.simulator.update_scheduler(new_scheduler)
        self.assertIsInstance(self.simulator.scheduler, UncontrolledCharging)
        self.assertIsInstance(new_scheduler.interface, Interface)
        self.assertEqual(self.simulator.max_recompute, 1)


class TestSimulatorWarnings(TestCase):
    def test_update_schedules_infeasible_schedule(self):
        network = ChargingNetwork()
        network.register_evse(EVSE("PS-001"), 240, 0)
        network.register_evse(EVSE("PS-004"), 240, 0)
        network.register_evse(EVSE("PS-003"), 240, 0)
        network.register_evse(EVSE("PS-002"), 240, 0)
        network.register_evse(EVSE("PS-006"), 240, 0)
        curr_dict1 = {"PS-001": 0.25, "PS-002": 0.50, "PS-003": -0.25}
        current1 = Current(curr_dict1)
        curr_dict2 = {"PS-006": 0.30, "PS-004": -0.60, "PS-002": 0.50}
        current2 = Current(curr_dict2)
        network.add_constraint(current1, 50, name="first_constraint")
        network.add_constraint(current2, 10)
        start = Mock(datetime)
        scheduler = create_autospec(BaseAlgorithm)
        scheduler.max_recompute = None
        events = EventQueue(events=[Event(1), Event(2)])
        simulator = Simulator(network, scheduler, events, start)

        bad_schedule = {
            "PS-001": [200, 0, 160, 0],
            "PS-004": [0, 0, 0, 0],
            "PS-003": [0, 0, 0, 0],
            "PS-002": [0, 0, 26, 0],
            "PS-006": [0, 0, 0, 40],
        }
        with self.assertWarnsRegex(
            UserWarning,
            r"Invalid schedule provided at iteration 0. "
            r"Max violation is 2.9999\d+? A on _const_1 at time index 2.",
        ):
            simulator._update_schedules(bad_schedule)



# coding=utf-8
""" Tests for Preprocessing algorithms. """
from unittest import TestCase
from unittest.mock import Mock
from numpy import testing as nptest
from acnportal.algorithms.preprocessing import *
from acnportal.algorithms.tests.generate_test_cases import *
from acnportal.acnsim.interface import SessionInfo, InfrastructureInfo

ARRIVAL_TIME = 0
SESSION_DUR = 5
ENERGY_DEMAND = 32 * 5
N = 3
CURRENT_TIME = 0
PERIOD = 5


class TestEnforcePilotLimit(TestCase):
    def test_pilot_greater_than_existing_max(
        self,
    ) -> None:  # pylint: disable=no-self-use
        sessions = session_generator(
            num_sessions=N,
            arrivals=[ARRIVAL_TIME] * N,
            departures=[ARRIVAL_TIME + SESSION_DUR] * N,
            requested_energy=[3.3] * N,
            remaining_energy=[3.3] * N,
            max_rates=[np.repeat(16, SESSION_DUR)] * N,
        )
        sessions = [SessionInfo(**s) for s in sessions]
        infrastructure = InfrastructureInfo(
            **single_phase_single_constraint(num_evses=N, limit=32)
        )
        modified_sessions = enforce_pilot_limit(sessions, infrastructure)
        for session in modified_sessions:
            nptest.assert_almost_equal(session.max_rates, 16)

    def test_pilot_less_than_existing_max(self) -> None:  # pylint: disable=no-self-use
        sessions = session_generator(
            num_sessions=N,
            arrivals=[ARRIVAL_TIME] * N,
            departures=[ARRIVAL_TIME + SESSION_DUR] * N,
            requested_energy=[3.3] * N,
            remaining_energy=[3.3] * N,
            max_rates=[np.repeat(40, SESSION_DUR)] * N,
        )
        sessions = [SessionInfo(**s) for s in sessions]
        infrastructure = InfrastructureInfo(
            **single_phase_single_constraint(num_evses=N, limit=32)
        )
        modified_sessions = enforce_pilot_limit(sessions, infrastructure)
        for session in modified_sessions:
            nptest.assert_almost_equal(session.max_rates, 32)


class TestReconcileMaxMin(TestCase):
    @staticmethod
    def _session_generation_helper(
        max_rates: int, choose_min: bool = True
    ) -> SessionInfo:
        session = SessionInfo(
            station_id="1",
            session_id="1",
            requested_energy=3.3,
            energy_delivered=0,
            arrival=0,
            departure=5,
            max_rates=max_rates,
            min_rates=8,
        )
        return reconcile_max_and_min(session, choose_min=choose_min)

    def test_max_greater_than_min(self) -> None:  # pylint: disable=no-self-use
        modified_session: SessionInfo = self._session_generation_helper(32)
        nptest.assert_almost_equal(modified_session.max_rates, 32)
        nptest.assert_almost_equal(modified_session.min_rates, 8)

    def test_max_equal_min(self) -> None:  # pylint: disable=no-self-use
        modified_session: SessionInfo = self._session_generation_helper(8)
        nptest.assert_almost_equal(modified_session.max_rates, 8)
        nptest.assert_almost_equal(modified_session.min_rates, 8)

    def test_max_less_than_min(self) -> None:  # pylint: disable=no-self-use
        modified_session: SessionInfo = self._session_generation_helper(6)
        nptest.assert_almost_equal(modified_session.max_rates, 8)
        nptest.assert_almost_equal(modified_session.min_rates, 8)

    def test_max_less_than_min_use_max(self) -> None:  # pylint: disable=no-self-use
        modified_session: SessionInfo = self._session_generation_helper(
            6, choose_min=False
        )
        nptest.assert_almost_equal(modified_session.max_rates, 6)
        nptest.assert_almost_equal(modified_session.min_rates, 6)


class TestExpandMaxMinRates(TestCase):
    def setUp(self) -> None:
        """
        Tests expand_max_min_rates in both the scalar and array max/min rate cases.
        """
        self.max_rates = [16, 24, 32]
        self.min_rates = [0, 4, 8]

    @staticmethod
    def _session_generation_helper(
        max_rates: List[Union[float, np.ndarray]],
        min_rates: List[Union[float, np.ndarray]],
    ) -> List[SessionInfo]:
        sessions: List[SessionDict] = session_generator(
            num_sessions=N,
            arrivals=[ARRIVAL_TIME] * N,
            departures=[ARRIVAL_TIME + SESSION_DUR] * N,
            requested_energy=[3.3] * N,
            remaining_energy=[3.3] * N,
            max_rates=max_rates,
            min_rates=min_rates,
        )
        sessions: List[SessionInfo] = [SessionInfo(**s) for s in sessions]
        # Since we want to test the expand function but the SessionInfo constructor
        # already expands the max/min rates, we manually set the max/min rates back to
        # the input max_rates/min_rates (possibly scalars) after construction.

        def _set_bounds(
            session: SessionInfo, max_rate: float, min_rate: float
        ) -> SessionInfo:
            session.max_rates = max_rate
            session.min_rates = min_rate
            return session

        sessions: List[SessionInfo] = [
            _set_bounds(session, max_rates[i], min_rates[i])
            for i, session in enumerate(sessions)
        ]
        return expand_max_min_rates(sessions)

    def _verify_rate_array(
        self,
        modified_sessions: List[SessionInfo],
        sess_idx: int,
        max_rate: float,
        min_rate: float,
        shape: int = SESSION_DUR,
    ):
        nptest.assert_almost_equal(modified_sessions[sess_idx].max_rates, max_rate)
        self.assertEqual(modified_sessions[sess_idx].max_rates.shape, (shape,))
        nptest.assert_almost_equal(modified_sessions[sess_idx].min_rates, min_rate)
        self.assertEqual(modified_sessions[sess_idx].min_rates.shape, (shape,))

    def test_expand_rates_arrays(self) -> None:
        max_rate_array: List[np.ndarray] = [
            np.array([rate] * SESSION_DUR) for rate in self.max_rates
        ]
        min_rate_array: List[np.ndarray] = [
            np.array([rate] * SESSION_DUR) for rate in self.min_rates
        ]
        modified_sessions = self._session_generation_helper(
            max_rates=max_rate_array, min_rates=min_rate_array,
        )
        self._verify_rate_array(modified_sessions, 0, 16, 0)
        self._verify_rate_array(modified_sessions, 1, 24, 4)
        self._verify_rate_array(modified_sessions, 2, 32, 8)

    def test_lower_existing_max_scalars(self) -> None:
        modified_sessions = self._session_generation_helper(
            max_rates=[16, 24, 32], min_rates=[0, 4, 8]
        )
        self._verify_rate_array(modified_sessions, 0, 16, 0)
        self._verify_rate_array(modified_sessions, 1, 24, 4)
        self._verify_rate_array(modified_sessions, 2, 32, 8)


class TestApplyUpperBoundEstimate(TestCase):
    @staticmethod
    def _session_generation_helper(
        max_rate_list: List[Union[float, np.ndarray]],
        upper_bound_estimate: Dict[str, float],
        expected_max: float,
        expected_min: float,
        min_rate_list: Optional[List[Union[float, np.ndarray]]] = None,
    ) -> None:
        if min_rate_list is not None:
            min_rate_list *= N
        sessions: List[SessionDict] = session_generator(
            num_sessions=N,
            arrivals=[ARRIVAL_TIME] * N,
            departures=[ARRIVAL_TIME + SESSION_DUR] * N,
            requested_energy=[3.3] * N,
            remaining_energy=[3.3] * N,
            max_rates=max_rate_list * N,
            min_rates=min_rate_list,
        )
        sessions: List[SessionInfo] = [SessionInfo(**s) for s in sessions]
        rd = Mock()
        rd.get_maximum_rates = Mock(return_value=upper_bound_estimate)
        modified_sessions = apply_upper_bound_estimate(rd, sessions)
        for session in modified_sessions:
            nptest.assert_almost_equal(session.max_rates, expected_max)
            nptest.assert_almost_equal(session.min_rates, expected_min)

    def test_default_max_rates_scalars(self) -> None:  # pylint: disable=no-self-use
        self._session_generation_helper(
            max_rate_list=[32],
            upper_bound_estimate={f"{i}": 16 for i in range(N)},
            expected_max=16,
            expected_min=0,
        )

    def test_lower_existing_max_scalars(self) -> None:  # pylint: disable=no-self-use
        self._session_generation_helper(
            max_rate_list=[12],
            upper_bound_estimate={f"{i}": 16 for i in range(N)},
            expected_max=12,
            expected_min=0,
        )

    def test_vector_existing_max_scalar_rampdown(
        self,
    ) -> None:  # pylint: disable=no-self-use
        self._session_generation_helper(
            max_rate_list=[np.repeat(32, SESSION_DUR)],
            upper_bound_estimate={f"{i}": 16 for i in range(N)},
            expected_max=16,
            expected_min=0,
        )

    def test_vector_lower_existing_max_scalar_rampdown(
        self,
    ) -> None:  # pylint: disable=no-self-use
        self._session_generation_helper(
            max_rate_list=[np.repeat(12, SESSION_DUR)],
            upper_bound_estimate={f"{i}": 16 for i in range(N)},
            expected_max=12,
            expected_min=0,
        )

    def test_all_vectors_rampdown_lower(self) -> None:  # pylint: disable=no-self-use
        self._session_generation_helper(
            max_rate_list=[np.repeat(32, SESSION_DUR)],
            upper_bound_estimate={f"{i}": [16] * 5 for i in range(N)},
            expected_max=16,
            expected_min=0,
        )

    def test_all_vectors_existing_lower(self) -> None:  # pylint: disable=no-self-use
        self._session_generation_helper(
            max_rate_list=[np.repeat(12, SESSION_DUR)],
            upper_bound_estimate={f"{i}": [16] * 5 for i in range(N)},
            expected_max=12,
            expected_min=0,
        )

    def test_minimum_rates_binding(self) -> None:  # pylint: disable=no-self-use
        self._session_generation_helper(
            max_rate_list=[np.repeat(12, SESSION_DUR)],
            upper_bound_estimate={f"{i}": 6 for i in range(N)},
            expected_max=8,
            expected_min=8,
            min_rate_list=[np.repeat(8, SESSION_DUR)],
        )

    def test_partial_estimator(self) -> None:  # pylint: disable=no-self-use
        sessions: List[SessionDict] = session_generator(
            num_sessions=N,
            arrivals=[ARRIVAL_TIME] * N,
            departures=[ARRIVAL_TIME + SESSION_DUR] * N,
            requested_energy=[3.3] * N,
            remaining_energy=[3.3] * N,
            max_rates=[32] * N,
        )
        sessions: List[SessionInfo] = [SessionInfo(**s) for s in sessions]
        rd = Mock()
        rd.get_maximum_rates = Mock(
            return_value={f"{i}": [16] * 5 for i in range(N) if i != 1}
        )
        modified_sessions = apply_upper_bound_estimate(rd, sessions)
        for i, session in enumerate(modified_sessions):
            nptest.assert_almost_equal(session.max_rates, (16 if i != 1 else 32))
            nptest.assert_almost_equal(session.min_rates, 0)


class TestApplyMinimumChargingRate(TestCase):
    @staticmethod
    def _session_generation_helper(
        max_rate_list: List[Union[float, np.ndarray]],
        min_rate_list: Optional[List[Union[float, np.ndarray]]] = None,
        remaining_energy: float = 3.3,
    ) -> List[SessionInfo]:
        if min_rate_list is not None:
            min_rate_list *= N
        sessions: List[SessionDict] = session_generator(
            num_sessions=N,
            arrivals=[ARRIVAL_TIME] * N,
            departures=[ARRIVAL_TIME + SESSION_DUR] * N,
            requested_energy=[3.3] * N,
            remaining_energy=[remaining_energy] * N,
            max_rates=max_rate_list * N,
            min_rates=min_rate_list,
        )
        sessions: List[SessionInfo] = [SessionInfo(**s) for s in sessions]
        infrastructure = InfrastructureInfo(**single_phase_single_constraint(N, 32))
        modified_sessions = apply_minimum_charging_rate(
            sessions, infrastructure, PERIOD
        )
        return modified_sessions

    def test_evse_less_than_session_max(self) -> None:  # pylint: disable=no-self-use
        modified_sessions: List[SessionInfo] = self._session_generation_helper(
            max_rate_list=[np.repeat(32, SESSION_DUR)]
        )
        for session in modified_sessions:
            nptest.assert_almost_equal(session.max_rates, 32)
            nptest.assert_almost_equal(session.min_rates[0], 8)
            nptest.assert_almost_equal(session.min_rates[1:], 0)

    def test_evse_less_than_existing_min(self) -> None:  # pylint: disable=no-self-use
        modified_sessions: List[SessionInfo] = self._session_generation_helper(
            max_rate_list=[np.repeat(32, SESSION_DUR)],
            min_rate_list=[np.repeat(16, SESSION_DUR)],
        )
        for session in modified_sessions:
            nptest.assert_almost_equal(session.max_rates, 32)
            nptest.assert_almost_equal(session.min_rates, 16)

    def test_evse_min_greater_than_remaining_energy(
        self,
    ) -> None:  # pylint: disable=no-self-use
        modified_sessions: List[SessionInfo] = self._session_generation_helper(
            max_rate_list=[np.repeat(32, SESSION_DUR)], remaining_energy=0.05,
        )
        for session in modified_sessions:
            nptest.assert_almost_equal(session.max_rates[0], 0)
            nptest.assert_almost_equal(session.max_rates[1:], 32)
            nptest.assert_almost_equal(session.min_rates[0], 0)
            nptest.assert_almost_equal(session.min_rates[1:], 0)

    def test_evse_min_greater_than_session_max(
        self,
    ) -> None:  # pylint: disable=no-self-use
        modified_sessions: List[SessionInfo] = self._session_generation_helper(
            max_rate_list=[np.repeat(6, SESSION_DUR)],
        )
        for session in modified_sessions:
            nptest.assert_almost_equal(session.max_rates[0], 8)
            nptest.assert_almost_equal(session.max_rates[1:], 6)
            nptest.assert_almost_equal(session.min_rates[0], 8)
            nptest.assert_almost_equal(session.min_rates[1:], 0)

    def test_apply_min_infeasible(self) -> None:  # pylint: disable=no-self-use
        sessions = session_generator(
            num_sessions=N,
            arrivals=[1, 2, 3],
            departures=[1 + SESSION_DUR, 2 + SESSION_DUR, 3 + SESSION_DUR],
            requested_energy=[3.3] * N,
            remaining_energy=[3.3] * N,
            max_rates=[np.repeat(32, SESSION_DUR)] * N,
        )
        sessions = [SessionInfo(**s) for s in sessions]
        infrastructure = InfrastructureInfo(**single_phase_single_constraint(N, 16))
        modified_sessions = apply_minimum_charging_rate(
            sessions, infrastructure, PERIOD
        )
        for i in range(2):
            nptest.assert_almost_equal(modified_sessions[i].min_rates[0], 8)
            nptest.assert_almost_equal(modified_sessions[i].min_rates[1:], 0)
        # It is not feasible to deliver 8 A to session '2', so max and
        # min should be 0 at time t=0.
        nptest.assert_almost_equal(modified_sessions[2].min_rates, 0)
        nptest.assert_almost_equal(modified_sessions[2].max_rates[0], 0)


class TestRemoveFinishedSessions(TestCase):
    def _test_remove_sessions_within_threshold(
        self, remaining_energies: List[float]
    ) -> None:
        sessions = session_generator(
            num_sessions=N,
            arrivals=[1, 2, 3],
            departures=[1 + SESSION_DUR, 2 + SESSION_DUR, 3 + SESSION_DUR],
            requested_energy=[3.3] * N,
            remaining_energy=remaining_energies,
            max_rates=[np.repeat(32, SESSION_DUR)] * N,
        )
        infrastructure = InfrastructureInfo(
            **three_phase_balanced_network(1, limit=100)
        )
        sessions = [SessionInfo(**s) for s in sessions]
        modified_sessions = remove_finished_sessions(sessions, infrastructure, 5)
        self.assertEqual(len(modified_sessions), 2)

    def test_remove_sessions_zero_remaining(self) -> None:
        self._test_remove_sessions_within_threshold([0, 3.3, 2])

    def test_remove_sessions_remaining_within_threshold(self) -> None:
        self._test_remove_sessions_within_threshold([0.1, 3.3, 2])



from .simulator import Simulator
from .interface import Interface, InvalidScheduleError

from .analysis import *
from .events import *
from .models import *
from .network import *

del simulator
del interface



from unittest import TestCase
from unittest.mock import create_autospec

from acnportal.acnsim.models import EV
from acnportal.acnsim.models import EVSE, DeadbandEVSE, FiniteRatesEVSE
from acnportal.acnsim.models import InvalidRateError, StationOccupiedError


class TestEVSE(TestCase):
    def setUp(self):
        self.max_rate = 32
        self.mid_rate = 16
        self.atol = 1e-3
        self.evse = EVSE("0001", max_rate=self.max_rate)

    def test_allowable_pilot_signals_default(self):
        inf_evse = EVSE("0001")
        self.assertEqual(inf_evse.allowable_pilot_signals, [0, float("inf")])

    def test_allowable_pilot_signals(self):
        self.assertEqual(self.evse.allowable_pilot_signals, [0, self.max_rate])

    def test_plugin_unoccupied(self):
        ev = create_autospec(EV)
        self.evse.plugin(ev)
        self.assertEqual(self.evse.ev, ev)

    def test_plugin_occupied(self):
        ev = create_autospec(EV)
        ev2 = create_autospec(EV)
        self.evse.plugin(ev)
        self.assertEqual(self.evse.ev, ev)
        with self.assertRaises(StationOccupiedError):
            self.evse.plugin(ev2)

    def test_unplug_occupied(self):
        ev = create_autospec(EV)
        self.evse.plugin(ev)
        self.evse.unplug()
        self.assertIsNone(self.evse.ev)

    def test_unplug_unoccupied(self):
        self.evse.unplug()
        self.assertIsNone(self.evse.ev)

    def _set_pilot_has_ev_valid_rate_helper(self, pilot):
        ev = create_autospec(EV)
        self.evse.plugin(ev)
        self.evse.set_pilot(pilot, 240, 5)
        self.assertEqual(self.evse.current_pilot, pilot)
        self.evse.ev.charge.assert_called_once()

    def test_set_pilot_has_ev_valid_rate(self):
        self._set_pilot_has_ev_valid_rate_helper(self.mid_rate)

    def test_set_pilot_has_ev_almost_large_rate(self):
        self._set_pilot_has_ev_valid_rate_helper(
            self.max_rate + self.atol - self.atol / 1e2
        )

    def test_set_pilot_has_ev_almost_small_rate(self):
        self._set_pilot_has_ev_valid_rate_helper(-self.atol + self.atol / 1e2)

    def test_set_pilot_has_ev_negative_rate(self):
        ev = create_autospec(EV)
        self.evse.plugin(ev)
        with self.assertRaises(InvalidRateError):
            self.evse.set_pilot(-1, 240, 5)

    def _set_pilot_no_ev_invalid_rate_helper(self, pilot):
        with self.assertRaises(InvalidRateError):
            self.evse.set_pilot(pilot, 240, 5)

    def test_set_pilot_no_ev_negative_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(-1)

    def test_set_pilot_no_ev_large_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(self.max_rate + 1)

    def test_set_pilot_no_ev_barely_large_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(
            self.max_rate + self.atol + self.atol / 1e2
        )

    def test_set_pilot_no_ev_barely_small_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(-self.atol - self.atol / 1e2)


class TestDeadbandEVSE(TestEVSE):
    def setUp(self):
        super().setUp()
        self.deadband_end = 6
        self.evse = DeadbandEVSE("0001", self.deadband_end, max_rate=self.max_rate)

    def test_allowable_pilot_signals(self):
        self.assertEqual(
            self.evse.allowable_pilot_signals, [self.deadband_end, self.max_rate]
        )

    def test_set_pilot_has_ev_almost_over_zero_rate(self):
        self._set_pilot_has_ev_valid_rate_helper(self.atol - self.atol / 1e2)

    def test_set_pilot_has_ev_almost_under_deadband_rate(self):
        self._set_pilot_has_ev_valid_rate_helper(
            self.deadband_end - self.atol + self.atol / 1e2
        )

    def test_set_pilot_has_ev_invalid_rate(self):
        ev = create_autospec(EV)
        self.evse.plugin(ev)
        with self.assertRaises(InvalidRateError):
            self.evse.set_pilot(self.deadband_end - 1, 240, 5)

    def test_set_pilot_no_ev_invalid_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(self.deadband_end - 1)

    def test_set_pilot_no_ev_barely_over_zero_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(self.atol + self.atol / 1e2)

    def test_set_pilot_has_ev_barely_under_deadband_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(
            self.deadband_end - self.atol - self.atol / 1e2
        )


class TestFiniteRatesEVSE(TestEVSE):
    def setUp(self):
        super().setUp()
        self.evse = FiniteRatesEVSE("0001", [0, 8, 16, 24, self.max_rate])

    def test_max_rate_finite_rates_evse(self):
        self.assertEqual(self.evse.max_rate, self.max_rate)

    def test_min_rate_finite_rates_evse(self):
        self.assertEqual(self.evse.min_rate, 8)

    def test_allowable_pilot_signals(self):
        self.assertEqual(
            self.evse.allowable_pilot_signals, [0, 8, 16, 24, self.max_rate]
        )

    def test_allowable_pilot_signals_unsorted(self):
        evse2 = FiniteRatesEVSE("0002", [self.max_rate, 16, 8, 24, 0])
        self.assertEqual(evse2.allowable_pilot_signals, [0, 8, 16, 24, self.max_rate])

    def test_allowable_pilot_signals_duplicates(self):
        evse3 = FiniteRatesEVSE("0003", [0, 8, 8, 24, 16, 24, self.max_rate])
        self.assertEqual(evse3.allowable_pilot_signals, [0, 8, 16, 24, self.max_rate])

    def test_allowable_pilot_signals_no_zero(self):
        evse4 = FiniteRatesEVSE("0004", [8, 24, 16, self.max_rate])
        self.assertEqual(evse4.allowable_pilot_signals, [0, 8, 16, 24, self.max_rate])

    def test_set_pilot_has_ev_almost_over_finite_rate(self):
        self._set_pilot_has_ev_valid_rate_helper(
            self.mid_rate + self.atol - self.atol / 1e2
        )

    def test_set_pilot_has_ev_almost_under_finite_rate(self):
        self._set_pilot_has_ev_valid_rate_helper(
            self.mid_rate - self.atol + self.atol / 1e2
        )

    def test_set_pilot_has_ev_invalid_rate(self):
        ev = create_autospec(EV)
        self.evse.plugin(ev)
        with self.assertRaises(InvalidRateError):
            self.evse.set_pilot(17.2, 240, 5)

    def test_set_pilot_no_ev_invalid_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(17.2)

    def test_set_pilot_no_ev_barely_over_finite_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(
            self.max_rate + self.atol + self.atol / 1e2
        )

    def test_set_pilot_no_ev_barely_under_finite_rate(self):
        self._set_pilot_no_ev_invalid_rate_helper(
            self.max_rate - self.atol - self.atol / 1e2
        )



from typing import Union, Dict, SupportsFloat, List, Optional

import pandas as pd


class Current(pd.Series):
    """ A simple representation of currents as an extension of pandas Series.
    Includes addition, subtraction, and multiplication (by scalar) operators.

    Attributes:
        loads (Dict[str, number]): Dictionary which maps a load_id to its coefficient in
        the aggregate current.

    Args:
        loads (Dict[str, number], str, or List[str], pd.Series): If dict, a dictionary
            mapping load_ids to coefficients. If str a load_id. If list, a list of
            load_ids. Default None. If None, loads will begin as an empty dict.
    """

    def __init__(
        self,
        loads: Optional[
            Union[Dict[str, SupportsFloat], str, List[str], pd.Series]
        ] = None,
    ):

        # Backwards compatibility with previous Current specification methods
        if isinstance(loads, dict):
            super().__init__(loads)
        elif isinstance(loads, str):
            super().__init__({loads: 1})
        elif loads is not None and all(isinstance(load, str) for load in loads):
            super().__init__({load_id: 1 for load_id in loads})
        elif isinstance(loads, pd.Series):
            super().__init__(loads)
        elif loads is None:
            # The object type is specified explicitly here to address a
            # warning in pandas 1.x.
            super().__init__(dtype="float64")
        else:
            raise TypeError(
                "Variable loads should be of type dict, str, pd.Series, or Lst[str]."
            )

    def __add__(self, other):
        """ Return new Current which is the sum of self and other.

        See add_current for description. Major difference is that __add__ returns a new Current object.

        Returns:
            Current: self + other
        Raises:
            TypeError: Raised if other is not of type Current.
        """
        if isinstance(other, Current):
            return Current(self.add(other, fill_value=0))
        else:
            TypeError("Must be of type Current.")

    # Allow for right addition as well.
    __radd__ = __add__

    def __sub__(self, other):
        """ Return Current which is self minus other.

        Args:
            other (Current): Current to be subtracted from self.
        Returns:
            Current: self - other
        """
        return Current(self.add(-1 * other, fill_value=0))



import numpy as np

from acnportal.acnsim.interface import InfrastructureInfo, SessionInfo


def infrastructure_constraints_feasible(
    rates: np.ndarray,
    infrastructure: InfrastructureInfo,
    linear: bool = False,
    violation_tolerance: float = 1e-5,
    relative_tolerance: float = 1e-7,
) -> bool:
    """ Return if a set of current magnitudes for each load are feasible.

    For a given constraint, the larger of the violation_tolerance
    and relative_tolerance is used to evaluate feasibility.

    This is a static version of acnsim.ChargingNetwork.is_feasible.

    Args:
        rates (np.ndarray): 2-D matrix with each row corresponding to
            an EVSE and each column corresponding to a time index in the schedule.
        infrastructure (InfrastructureInfo): The InfrastructureInfo object that contains
            information about the network constraints.
        linear (bool): If True, linearize all constraints to a more conservative
            but easier to compute constraint by ignoring the phase angle and taking
            the absolute value of all load coefficients. Default False.
        violation_tolerance (float): Absolute amount by which
            rates may violate network constraints. Default 1e-5
        relative_tolerance (float): Relative amount by which
            schedule_matrix may violate network constraints. Default 1e-7

    Returns:
        bool: If load_currents is feasible according to this set of
            constraints.
    """
    tol = np.maximum(
        violation_tolerance, relative_tolerance * infrastructure.constraint_limits
    )
    if not linear:
        phase_in_rad = np.deg2rad(infrastructure.phases)
        for j, v in enumerate(infrastructure.constraint_matrix):
            a = np.stack([v * np.cos(phase_in_rad), v * np.sin(phase_in_rad)])
            line_currents = np.linalg.norm(a @ rates, axis=0)
            if not np.all(
                line_currents <= infrastructure.constraint_limits[j] + tol[j]
            ):
                return False
    else:
        for j, v in enumerate(infrastructure.constraint_matrix):
            line_currents = np.linalg.norm(np.abs(v) @ rates, axis=0)
            if not np.all(
                line_currents <= infrastructure.constraint_limits[j] + tol[j]
            ):
                return False
    return True


def remaining_amp_periods(
    session: SessionInfo, infrastructure: InfrastructureInfo, period: float
) -> float:
    """ Return the session's remaining demand in A*periods. This function is a static
    version of acnsim.Interface.remaining_amp_periods.

    Args:
        session (SessionInfo): The SessionInfo object for which to get remaining demand.
        infrastructure (InfrastructureInfo): The InfrastructureInfo object that contains
            voltage information about the network.
        period (float): Period of the simulation in minutes.

    Returns:
        float: the EV's remaining demand in A*periods.
    """
    i = infrastructure.get_station_index(session.station_id)
    amp_hours = session.remaining_demand * 1000 / infrastructure.voltages[i]
    return amp_hours * 60 / period



# coding=utf-8
"""
Sorting-based scheduling algorithms.
"""
from collections import deque
from copy import copy
from typing import Callable, List, Optional, Dict

import numpy as np

from .upper_bound_estimator import UpperBoundEstimatorBase
from .base_algorithm import BaseAlgorithm
from .utils import infrastructure_constraints_feasible
from .postprocessing import format_array_schedule
from .preprocessing import (
    enforce_pilot_limit,
    apply_upper_bound_estimate,
    apply_minimum_charging_rate,
    remove_finished_sessions,
)
from warnings import warn

from acnportal.acnsim.interface import SessionInfo, InfrastructureInfo, Interface


class SortedSchedulingAlgo(BaseAlgorithm):
    """ Class for sorting based algorithms like First Come First Served (FCFS) and
    Earliest Deadline First (EDF).

    Implements abstract class BaseAlgorithm.

    For this family of algorithms, active EVs are first sorted by some metric, then
    current is allocated to each EV in order. To allocate current we use a binary
    search approach which allocates each EV the maximum current possible subject to the
    constraints and already allocated allotments.

    The argument sort_fn controlled how the EVs are sorted and thus which sorting based
    algorithm is implemented.

    Args:
        sort_fn (Callable[[List[SessionInfo], Interface], List[SessionInfo]]): Function
            which takes in a list of SessionInfo objects and returns a list of the
            same SessionInfo objects but sorted according to some metric.
    """

    _sort_fn: Callable[[List[SessionInfo], Interface], List[SessionInfo]]
    estimate_max_rate: bool
    max_rate_estimator: Optional[UpperBoundEstimatorBase]
    uninterrupted_charging: bool
    allow_overcharging: bool

    def __init__(
        self,
        sort_fn: Callable[[List[SessionInfo], Interface], List[SessionInfo]],
        estimate_max_rate: bool = False,
        max_rate_estimator: Optional[UpperBoundEstimatorBase] = None,
        uninterrupted_charging: bool = False,
        allow_overcharging: bool = False,
    ) -> None:
        super().__init__()
        self._sort_fn = sort_fn
        # Call algorithm each period since it only returns a rate for the
        # next period.
        self.max_recompute = 1
        self.estimate_max_rate = estimate_max_rate
        self.max_rate_estimator = max_rate_estimator
        self.uninterrupted_charging = uninterrupted_charging
        self.allow_overcharging = allow_overcharging

    def register_interface(self, interface: Interface) -> None:
        """ Register interface to the _simulator/physical system.

        This interface is the only connection between the algorithm and what it
            is controlling. Its purpose is to abstract the underlying
            network so that the same algorithms can run on a simulated
            environment or a physical one.

        Args:
            interface (Interface): An interface to the underlying network
                whether simulated or real.

        Returns:
            None
        """
        self._interface = interface
        if self.max_rate_estimator is not None:
            self.max_rate_estimator.register_interface(interface)

    def run_preprocessing(
        self, active_sessions: List[SessionInfo], infrastructure: InfrastructureInfo
    ) -> List[SessionInfo]:
        """ Run a set of preprocessing functions on the active_sessions given to the
        algorithm.

        Args:
            active_sessions (List[SessionInfo]): see BaseAlgorithm
            infrastructure (InfrastructureInfo): Description of the electrical
                infrastructure.

        Returns:
            List[SessionInfo]: A list of processed SessionInfo objects.

        """
        if self.allow_overcharging:
            warn(
                "allow_overcharging is currently not supported. It will be added in a "
                "future release."
            )
        active_sessions: List[SessionInfo] = remove_finished_sessions(
            active_sessions, infrastructure, self.interface.period
        )
        active_sessions = enforce_pilot_limit(active_sessions, infrastructure)
        if self.estimate_max_rate:
            active_sessions: List[SessionInfo] = apply_upper_bound_estimate(
                self.max_rate_estimator, active_sessions
            )
        if self.uninterrupted_charging:
            active_sessions: List[SessionInfo] = apply_minimum_charging_rate(
                active_sessions, infrastructure, self.interface.period
            )
        return active_sessions

    def sorting_algorithm(
        self, active_sessions: List[SessionInfo], infrastructure: InfrastructureInfo
    ) -> np.ndarray:
        """ Schedule EVs by first sorting them by sort_fn, then allocating
            them their maximum feasible rate.

        See class documentation for description of the algorithm.

        Args:
            active_sessions (List[SessionInfo]): see BaseAlgorithm
            infrastructure (InfrastructureInfo): Description of the electrical
                infrastructure.

        Returns:
            np.array[Union[float, int]]: Array of charging rates, where each
                row is the charging rates for a specific session.
        """
        queue: List[SessionInfo] = self._sort_fn(active_sessions, self.interface)
        schedule: np.ndarray = np.zeros(infrastructure.num_stations)

        # Start each EV at its lower bound
        for session in queue:
            station_index: int = infrastructure.get_station_index(session.station_id)
            lb: float = max(0, session.min_rates[0])
            schedule[station_index] = lb

        if not infrastructure_constraints_feasible(schedule, infrastructure):
            raise ValueError(
                "Charging all sessions at their lower bound is not feasible."
            )

        for session in queue:
            station_index = infrastructure.get_station_index(session.station_id)
            ub: float = min(
                session.max_rates[0], self.interface.remaining_amp_periods(session)
            )
            lb: float = max(0, session.min_rates[0])
            if infrastructure.is_continuous[station_index]:
                charging_rate: float = self.max_feasible_rate(
                    station_index, ub, schedule, infrastructure, eps=0.01, lb=lb
                )
            else:
                allowable = [
                    a
                    for a in infrastructure.allowable_pilots[station_index]
                    if lb <= a <= ub
                ]

                if len(allowable) == 0:
                    charging_rate = 0
                else:
                    charging_rate = self.discrete_max_feasible_rate(
                        station_index, allowable, schedule, infrastructure
                    )
            schedule[station_index] = charging_rate
        return schedule

    @staticmethod
    def max_feasible_rate(
        station_index: int,
        ub: float,
        schedule: np.ndarray,
        infrastructure: InfrastructureInfo,
        eps: float = 0.0001,
        lb: float = 0.0,
    ) -> float:
        """ Return the maximum feasible rate less than ub subject to the environment's
        constraints.

        If schedule contains non-zero elements at the given time, these are
        treated as fixed allocations and this function will include them
        when determining the maximum feasible rate for the given EVSE.

        Args:
            station_index (int): Index for the station in the schedule
                vector.
            ub (float): Upper bound on the charging rate. [A]
            schedule (np.array[Union[float, int]]): Array of charging rates, where each
                row is the charging rates for a specific session.
            infrastructure (InfrastructureInfo): Description of the electrical
                infrastructure.
            eps (float): Accuracy to which the max rate should be calculated.
                (When the binary search is terminated.)
            lb (float): Lower bound on the charging rate [A]

        Returns:
            float: maximum feasible rate less than ub subject to the
                environment's constraints. [A]
        """

        def bisection(
            _index: int, _lb: float, _ub: float, _schedule: np.ndarray
        ) -> float:
            """ Use the bisection method to find the maximum feasible charging
                rate for the EV. """
            mid: float = (_ub + _lb) / 2
            _new_schedule = copy(schedule)
            _new_schedule[_index] = mid
            if (_ub - _lb) <= eps:
                return _lb
            elif infrastructure_constraints_feasible(_new_schedule, infrastructure):
                return bisection(_index, mid, _ub, _new_schedule)
            else:
                return bisection(_index, _lb, mid, _new_schedule)

        if not infrastructure_constraints_feasible(schedule, infrastructure):
            raise ValueError("The initial schedule is not feasible.")

        # Test maximum rate to short-circuit bisection
        new_schedule = copy(schedule)
        new_schedule[station_index] = ub
        if infrastructure_constraints_feasible(new_schedule, infrastructure):
            return ub
        return bisection(station_index, lb, ub, schedule)

    @staticmethod
    def discrete_max_feasible_rate(
        station_index: int,
        allowable_pilots: List[float],
        schedule: np.ndarray,
        infrastructure: InfrastructureInfo,
    ) -> float:
        """ Return the maximum feasible allowable rate subject to the
            infrastructure's constraints and the discrete pilot constraints of this
            station.

        If schedule contains non-zero elements at the given time, these are
        treated as fixed allocations and this function will include them
        when determining the maximum feasible rate for the given EVSE.

        Args:
            station_index (int): Index for the station in the schedule
                vector.
            allowable_pilots (List[float]): List of allowable charging rates
                sorted in ascending order.
            schedule (np.array[Union[float, int]]): Array of charging rates, where each
                row is the charging rates for a specific session.
            infrastructure (InfrastructureInfo): Description of the electrical
                infrastructure.

        Returns:
            float: maximum feasible rate less than ub subject to the
                infrastructure's constraints. [A]
        """
        if not infrastructure_constraints_feasible(schedule, infrastructure):
            raise ValueError("The initial schedule is not feasible.")
        new_schedule = copy(schedule)
        feasible_idx = len(allowable_pilots) - 1
        new_schedule[station_index] = allowable_pilots[feasible_idx]
        while not infrastructure_constraints_feasible(new_schedule, infrastructure):
            feasible_idx -= 1
            if feasible_idx < 0:
                new_schedule[station_index] = 0
                break
            else:
                new_schedule[station_index] = allowable_pilots[feasible_idx]
        return new_schedule[station_index]

    # noinspection PyMethodMayBeStatic
    def run_postprocessing(
        self, raw_schedule: np.ndarray, infrastructure: InfrastructureInfo
    ) -> Dict[str, List[float]]:
        """ Run a set of postprocessing functions on the schedule returned by the
        algorithm

        Args:
            raw_schedule (np.ndarray): An unprocessed schedule returned by a step of the
                algorithm.
            infrastructure (InfrastructureInfo): Description of the electrical
                infrastructure.

        Returns:
            Dict[str, List[float]]: Output schedule in a Simulator-accepted form (see
                BaseAlgorithm.schedule).

        """
        return format_array_schedule(raw_schedule, infrastructure)

    def schedule(self, active_sessions: List[SessionInfo]) -> Dict[str, List[float]]:
        """ Schedule EVs by first sorting them by sort_fn, then allocating them their
        maximum feasible rate.

        Implements abstract method schedule from BaseAlgorithm.

        See class documentation for description of the algorithm.

        Args:
            active_sessions (List[SessionInfo]): see BaseAlgorithm

        Returns:
            Dict[str, List[float]]: see BaseAlgorithm
        """
        infrastructure = self.interface.infrastructure_info()
        active_sessions = self.run_preprocessing(active_sessions, infrastructure)
        array_schedule = self.sorting_algorithm(active_sessions, infrastructure)
        return self.run_postprocessing(array_schedule, infrastructure)


class RoundRobin(SortedSchedulingAlgo):
    """ Family of algorithms which allocate charging rates among active EVs
        using a round robin approach.

    Extends SortingAlgorithm.

    For this family of algorithms EVs are first sorted as in SortingAlgorithm.
    The difference however, is that instead of allocating each EV its
    maximum charging rate as we go down the list, we instead give each EV
    one  unit of charge if it is feasible to do so. When it ceases to be
    feasible to give an EV more charge, it is removed from the list.
    This process continues until the list of EVs is empty.

    The argument sort_fn controlled how the EVs are sorted. This controls
    which  EVs will get potential higher charging rates when infrastructure
    constraints become binding.

    Args:
        sort_fn (Callable[[List[SessionInfo], Interface], List[SessionInfo]]): Function
            which takes in a list of EVs and returns a list of the same EVs but sorted
            according to some metric.
        continuous_inc (float): Increment to use when pilot signal is
            continuously controllable.
    """

    continuous_inc: float

    def __init__(
        self,
        sort_fn: Callable[[List[SessionInfo], Interface], List[SessionInfo]],
        estimate_max_rate: bool = False,
        max_rate_estimator: Optional[UpperBoundEstimatorBase] = None,
        uninterrupted_charging: bool = False,
        continuous_inc: float = 0.1,
        allow_overcharging: bool = False,
    ) -> None:
        super().__init__(
            sort_fn,
            estimate_max_rate,
            max_rate_estimator,
            uninterrupted_charging,
            allow_overcharging,
        )
        self.continuous_inc = continuous_inc

    def round_robin(
        self, active_sessions: List[SessionInfo], infrastructure: InfrastructureInfo
    ) -> np.ndarray:
        """ Schedule EVs using a round robin based equal sharing scheme.

        Implements abstract method schedule from BaseAlgorithm.

        See class documentation for description of the algorithm.

        Args:
            active_sessions (List[SessionInfo]): see BaseAlgorithm
            infrastructure (InfrastructureInfo): Description of electrical
                infrastructure.

        Returns:
            np.array[Union[float, int]]: Array of charging rates, where each
                row is the charging rates for a specific session.
        """
        queue = deque(self._sort_fn(active_sessions, self.interface))
        schedule = np.zeros(infrastructure.num_stations)
        rate_idx = np.zeros(infrastructure.num_stations, dtype=int)
        allowable_pilots = infrastructure.allowable_pilots.copy()
        for session in queue:
            i = infrastructure.get_station_index(session.station_id)
            # If pilot signal is continuous discretize it with increments of
            # continuous_inc.
            if infrastructure.is_continuous[i]:
                allowable_pilots[i] = np.arange(
                    session.min_rates[0],
                    session.max_rates[0] + self.continuous_inc / 2,
                    self.continuous_inc,
                )
            ub = min(
                session.max_rates[0],
                infrastructure.max_pilot[i],
                self.interface.remaining_amp_periods(session),
            )
            lb = max(0, session.min_rates[0])
            # Remove any charging rates which are not feasible.
            allowable_pilots[i] = allowable_pilots[i][lb <= allowable_pilots[i]]
            allowable_pilots[i] = allowable_pilots[i][allowable_pilots[i] <= ub]
            # All charging rates should start at their lower bound
            schedule[i] = allowable_pilots[i][0] if len(allowable_pilots[i]) > 0 else 0

        if not infrastructure_constraints_feasible(schedule, infrastructure):
            raise ValueError(
                "Charging all sessions at their lower bound is not feasible."
            )

        while len(queue) > 0:
            session = queue.popleft()
            i = infrastructure.get_station_index(session.station_id)
            if rate_idx[i] < len(allowable_pilots[i]) - 1:
                schedule[i] = allowable_pilots[i][rate_idx[i] + 1]
                if infrastructure_constraints_feasible(schedule, infrastructure):
                    rate_idx[i] += 1
                    queue.append(session)
                else:
                    schedule[i] = allowable_pilots[i][rate_idx[i]]
        return schedule

    def schedule(self, active_sessions: List[SessionInfo]) -> Dict[str, List[float]]:
        """ Schedule EVs by first sorting them by sort_fn, then allocating them their
        maximum feasible rate.

        Implements abstract method schedule from BaseAlgorithm.

        See class documentation for description of the algorithm.

        Args:
            active_sessions (List[SessionInfo]): see BaseAlgorithm

        Returns:
            Dict[str, List[float]]: see BaseAlgorithm
        """
        infrastructure = self.interface.infrastructure_info()
        active_sessions = self.run_preprocessing(active_sessions, infrastructure)
        array_schedule = self.round_robin(active_sessions, infrastructure)
        return self.run_postprocessing(array_schedule, infrastructure)


# -------------------- Sorting Functions --------------------------
# noinspection PyUnusedLocal
def first_come_first_served(
    evs: List[SessionInfo], iface: Interface
) -> List[SessionInfo]:
    """ Sort EVs by arrival time in increasing order.

    Args:
        evs (List[SessionInfo]): List of EVs to be sorted.
        iface (Interface): Interface object. (not used in this case)

    Returns:
        List[SessionInfo]: List of EVs sorted by arrival time in increasing order.
    """
    return sorted(evs, key=lambda x: x.arrival)


# noinspection PyUnusedLocal
def last_come_first_served(
    evs: List[SessionInfo], iface: Interface
) -> List[SessionInfo]:
    """ Sort EVs by arrival time in reverse order.
    Args:
       evs (List[SessionInfo]): List of EVs to be sorted.
       iface (Interface): Interface object. (not used in this case)
    Returns:
       List[SessionInfo]: List of EVs sorted by arrival time in decreasing order.
    """
    return sorted(evs, key=lambda x: x.arrival, reverse=True)


# noinspection PyUnusedLocal
def earliest_deadline_first(
    evs: List[SessionInfo], iface: Interface
) -> List[SessionInfo]:
    """ Sort EVs by estimated departure time in increasing order.

    Args:
        evs (List[SessionInfo]): List of EVs to be sorted.
        iface (Interface): Interface object. (not used in this case)

    Returns:
        List[SessionInfo]: List of EVs sorted by estimated departure time in increasing order.
    """
    return sorted(evs, key=lambda x: x.estimated_departure)


def least_laxity_first(evs: List[SessionInfo], iface: Interface) -> List[SessionInfo]:
    """ Sort EVs by laxity in increasing order.

    Laxity is a measure of the charging flexibility of an EV. Here we define laxity as:
        LAX_i(t) = (estimated_departure_i - t) - (remaining_demand_i(t) / max_rate_i)

    Args:
        evs (List[SessionInfo]): List of EVs to be sorted.
        iface (Interface): Interface object.

    Returns:
        List[SessionInfo]: List of EVs sorted by laxity in increasing order.
    """

    def laxity(ev: SessionInfo) -> float:
        """ Calculate laxity of the EV.

        Args:
            ev (EV): An EV object.

        Returns:
            float: The laxity of the EV.
        """
        lax = (ev.estimated_departure - iface.current_time) - (
            iface.remaining_amp_periods(ev) / iface.max_pilot_signal(ev.station_id)
        )
        return lax

    return sorted(evs, key=laxity)


def largest_remaining_processing_time(
    evs: List[SessionInfo], iface: Interface
) -> List[SessionInfo]:
    """ Sort EVs in decreasing order by the time taken to finish charging them at the
    EVSE's maximum rate.

    Args:
        evs (List[SessionInfo]): List of SessionInfo objects to be sorted.
        iface (Interface): Interface object.

    Returns:
        List[SessionInfo]: List of EVs sorted by remaining processing time in
            decreasing order.
    """

    def remaining_processing_time(ev: SessionInfo) -> float:
        """ Calculate minimum time needed to fully charge the EV based its remaining
        energy request and the EVSE's max charging rate.

        Args:
            ev (SessionInfo): A SessionInfo object.

        Returns:
            float: The minimum remaining processing time of the EV.
        """
        rpt = iface.remaining_amp_periods(ev) / iface.max_pilot_signal(ev.station_id)
        return rpt

    return sorted(evs, key=remaining_processing_time, reverse=True)



# coding=utf-8
"""
This module contains methods for accessing a static simulator mocked through a data
file.
"""
from typing import Dict, Any, List, Tuple, Optional
from warnings import warn

import numpy as np
from acnportal.acnsim.interface import (
    Interface,
    SessionInfo,
    InfrastructureInfo,
    Constraint,
)
from acnportal.algorithms import infrastructure_constraints_feasible


class TestingInterface(Interface):
    """ Static interface based on an attributes dict for testing algorithms without a
        full environment.
    """

    data: Dict[str, Any]

    def __init__(self, data: Dict[str, Any]) -> None:
        super().__init__(None)
        self.data = data

    def _get_or_error(self, field: str) -> Any:
        if field in self.data:
            return self.data[field]
        else:
            raise NotImplementedError(f"No data provided for {field}.")

    @property
    def last_applied_pilot_signals(self) -> Dict[str, float]:
        """ Return the pilot signals that were applied in the last _iteration of the
        simulation for all active EVs.

        Does not include EVs that arrived in the current _iteration.

        Returns:
            Dict[str, number]: A dictionary with the session ID as key and the pilot
                signal as value.
        """
        return self._get_or_error("last_applied_pilot_signals")

    @property
    def last_actual_charging_rate(self) -> Dict[str, float]:
        """ Return the actual charging rates in the last period for all active EVs.

        Returns:
            Dict[str, number]:  A dictionary with the session ID as key and actual
                charging rate as value.
        """
        return self._get_or_error("last_actual_charging_rate")

    @property
    def current_time(self) -> int:
        """ Get the current time (the current _iteration) of the simulator.

        Returns:
            int: The current _iteration of the simulator.
        """
        return self._get_or_error("current_time")

    @property
    def period(self) -> float:
        """ Return the length of each timestep in the simulation.

        Returns:
            int: Length of each time interval in the simulation. [minutes]
        """
        return self._get_or_error("period")

    def active_sessions(self) -> List[SessionInfo]:
        """ Return a list of SessionInfo objects describing the currently charging EVs.

        Returns:
            List[SessionInfo]: List of currently active charging sessions.
        """
        active_sessions = self._get_or_error("active_sessions")
        return [
            SessionInfo(current_time=self.current_time, **session)
            for session in active_sessions
        ]

    def infrastructure_info(self) -> InfrastructureInfo:
        """ Returns an InfrastructureInfo object generated from interface.

        Returns:
            InfrastructureInfo: A description of the charging infrastructure.
        """
        infrastructure = self._get_or_error("infrastructure_info")
        return InfrastructureInfo(
            np.array(infrastructure["constraint_matrix"]),
            np.array(infrastructure["constraint_limits"]),
            np.array(infrastructure["phases"]),
            np.array(infrastructure["voltages"]),
            infrastructure["constraint_ids"],
            infrastructure["station_ids"],
            np.array(infrastructure["max_pilot"]),
            np.array(infrastructure["min_pilot"]),
            [np.array(allowable) for allowable in infrastructure["allowable_pilots"]],
            np.array(infrastructure["is_continuous"]),
        )

    def allowable_pilot_signals(self, station_id: str) -> Tuple[bool, List[float]]:
        """ Returns the allowable pilot signal levels for the specified EVSE.
        One may assume an EVSE pilot signal of 0 is allowed regardless
        of this function's return values.

        Args:
            station_id (str): The ID of the station for which the allowable rates
                should be returned.

        Returns:
            bool: If the range is continuous or not
            list[float]: The sorted set of acceptable pilot signals. If continuous this
                range will have 2 values the min and the max acceptable values. [A]
        """
        infrastructure = self._get_or_error("infrastructure_info")
        i = infrastructure["station_ids"].index(station_id)
        return (
            infrastructure["is_continuous"][i],
            infrastructure["allowable_pilots"][i],
        )

    def max_pilot_signal(self, station_id: str) -> float:
        """ Returns the maximum allowable pilot signal level for the specified EVSE.

        Args:
            station_id (str): The ID of the station for which the allowable rates
                should be returned.

        Returns:
            float: the maximum pilot signal supported by this EVSE. [A]
        """
        infrastructure = self._get_or_error("infrastructure_info")
        i = infrastructure["station_ids"].index(station_id)
        return infrastructure["max_pilot"][i]

    def min_pilot_signal(self, station_id: str) -> float:
        """ Returns the minimum allowable pilot signal level for the specified EVSE.

        Args:
            station_id (str): The ID of the station for which the allowable rates
                should be returned.

        Returns:
            float: the minimum pilot signal supported by this EVSE. [A]
        """
        infrastructure = self._get_or_error("infrastructure_info")
        i = infrastructure["station_ids"].index(station_id)
        return infrastructure["min_pilot"][i]

    def evse_voltage(self, station_id: str) -> float:
        """ Returns the voltage of the EVSE.

        Args:
            station_id (str): The ID of the station for which the allowable rates
                should be returned.

        Returns:
            float: voltage of the EVSE. [V]
        """
        infrastructure = self._get_or_error("infrastructure_info")
        i = infrastructure["station_ids"].index(station_id)
        return infrastructure["voltages"][i]

    def evse_phase(self, station_id: str) -> float:
        """ Returns the phase angle of the EVSE.

        Args:
            station_id (str): The ID of the station for which the allowable rates
                should be returned.

        Returns:
            float: phase angle of the EVSE. [degrees]
        """
        infrastructure = self._get_or_error("infrastructure_info")
        i = infrastructure["station_ids"].index(station_id)
        return infrastructure["phases"][i]

    def get_constraints(self) -> Constraint:
        """ Get the constraint matrix and corresponding EVSE ids for the network.

        Returns:
            np.ndarray: Matrix representing the constraints of the network. Each row is
                a constraint and each
        """
        infrastructure = self._get_or_error("infrastructure_info")
        return Constraint(
            np.array(infrastructure["constraint_matrix"]),
            np.array(infrastructure["constraint_limits"]),
            infrastructure["constraint_ids"],
            infrastructure["station_ids"],
        )

    def is_feasible(
        self,
        load_currents: Dict[str, List[float]],
        linear: bool = False,
        violation_tolerance: float = 1e-5,
        relative_tolerance: float = 1e-7,
    ) -> bool:
        """ Return if a set of current magnitudes for each load are feasible.

        Wraps Network's is_feasible method.

        For a given constraint, the larger of the violation_tolerance
        and relative_tolerance is used to evaluate feasibility.

        Args:
            load_currents (Dict[str, List[number]]): Dictionary mapping load_ids to
                schedules of charging rates.
            linear (bool): If True, linearize all constraints to a more conservative
                but easier to compute constraint by ignoring the phase angle and taking
                the absolute value of all load coefficients. Default False.
            violation_tolerance (float): Absolute amount by which
                schedule may violate network constraints. Default
                None, in which case the network's violation_tolerance
                attribute is used.
            relative_tolerance (float): Relative amount by which
                schedule may violate network constraints. Default
                None, in which case the network's relative_tolerance
                attribute is used.

        Returns:
            bool: If load_currents is feasible at time t according to this set of
                constraints.
        """
        if len(load_currents) == 0:
            return True
        # Check that all schedules are the same length
        schedule_lengths = set(len(x) for x in load_currents.values())
        if len(schedule_lengths) > 1:
            raise InvalidScheduleError("All schedules should have the same length.")
        schedule_length = schedule_lengths.pop()

        # Convert input schedule into its matrix representation
        infrastructure = self._get_or_error("infrastructure_info")
        station_ids = infrastructure["station_ids"]
        schedule_matrix = np.array(
            [
                load_currents[station_id]
                if station_id in load_currents
                else [0] * schedule_length
                for station_id in station_ids
            ]
        )
        infrastructure = self.infrastructure_info()
        return infrastructure_constraints_feasible(
            schedule_matrix,
            infrastructure,
            linear,
            violation_tolerance,
            relative_tolerance,
        )

    def get_prices(self, length: int, start: Optional[int] = None) -> np.ndarray:
        """ Get a vector of prices beginning at time start and continuing for length
        periods. ($/kWh)

        Args:
            length (int): Number of elements in the prices vector. One entry per period.
            start (int): Time step of the simulation where price vector should begin.
                If None, uses the current timestep of the simulation. Default None.

        Returns:
            np.ndarray[float]: Array of floats where each entry is the price for the
                corresponding period. ($/kWh)
        """
        tariff = self._get_or_error("tariff")
        if start is None:
            start = self.current_time
        return np.array(tariff["tou_prices"][start : start + length])

    def get_demand_charge(self, start: Optional[int] = None) -> float:
        """ Get the demand charge for the given period. ($/kW)

        Args:
            start (int): Time step of the simulation where price vector should begin.
                If None, uses the current timestep of the simulation. Default None.

        Returns:
            float: Demand charge for the given period. ($/kW)
        """
        if start is not None:
            warn(
                "start time specified in get_demand_charge. "
                "JSONInterface currently only supports static demand charge"
            )
        tariff = self._get_or_error("tariff")
        return tariff["demand_charge"]

    def get_prev_peak(self) -> float:
        """ Get the highest aggregate peak demand so far in the simulation.

        Returns:
            float: Peak demand so far in the simulation. (A)
        """
        return self._get_or_error("prev_peak")


class InvalidScheduleError(Exception):
    """ Raised when the schedule passed to the simulator is invalid. """
